#!/bin/bash

TAG=10.0.2.2
VALFILE="../helm/values.yaml"
FIXDIR="../fixes"
BKUPDIR="../backup/$TAG"
LOGDIR="../logs"
FIXPKG=fixPackage.tgz
HISTFILE="../.installedFixes"
DRYRUN=0
LASTGOOD=""

print_help() {
	echo "Name: IMFixInstaller.sh"
	echo "When to run: Whenever a Fixpack or Interim Fix needs to be applied."
	echo "Description:"
	echo "   The latest version of this script should be downloaded from FixCentral"
	echo "   before running it.  It will perform the steps necessary to upgrade the"
	echo "   installation from it's current level to the level of the package"
	echo "   downloaded from FixCentral."
	echo "Options:"
	echo "   -dryrun: Will report on the actions that will be taken during the upgrade."
	echo "   -version: Will display the fixlevel of the starter area and the container."
	echo "   -history: Will display the history of fixes installed."
	echo "Example usage:"
	echo "   $ ./IMFixInstaller.sh"
	echo "   $ ./IMFixInstaller.sh -dryrun"
	echo "   $ ./IMFixINstaller.sh -version"
	echo ""
	exit 0
} #print_help

print_end() {
	log "**********                       **********"
	log "********** Ending IMFixInstaller **********"
	log "**********                       **********"
} #print_end

# Provide instructions on cleanup so we can leave it running while they
# investigate the error.
clean_up_info() {
	echo "Once the problem has been determined, you can reset the env with:"
	echo ""
	echo "$kubectl -n $NS delete deployment isvgimconfig"
	echo "$kubectl -n $NS scale --replicas=$REPLICAS sts isvgim"
	echo "rm -rf $FIXDIR"
} #clean_up_info

restore_tag() {
	printf "\nRestoring values.yaml file from backup\n"
	log "Restoring values.yaml file from backup"
	cp $BKUPDIR/values.yaml $VALFILE
	log "cp $BKUPDIR/values.yaml $VALFILE"
} #restore_tag 

# Convenience method to group together actions before any fixes are applied
early_reset() {
	clean_up_info
	restore_tag
	print_end
} #early_reset

# Called after we start applying fixes
finish_install() {
	printf "\nRemoving isvgimconfig pod"
	log "Removing isvgimconfig pod"
	$kubectl -n $NS delete deployment isvgimconfig
	log "$kubectl -n $NS delete deployment isvgimconfig"
	echo "Updating isvgim yaml"
	log "Transforming 300-statefulset-isvgim.yaml"
	./updateYaml.sh 300-statefulset-isvgim.yaml > /dev/null 2>&1
	log "./updateYaml.sh 300-statefulset-isvgim.yaml"
	$kubectl -n $NS scale --replicas=0 sts isvgim > /dev/null 2>&1
	log "$kubectl -n $NS scale --replicas=0 sts isvgim"
	printf "\n"
	read -p "Would you like to restart the IM server? (y/n) " YESNO
	printf "\n"
	if [ "x$YESNO" = "xy" ]; then
		log "User specified to re-start the IM server"
		if [ $REPLICAS -gt 0 ]; then
			printf "Restoring IM pods to $REPLICAS replicas.\n"
			log "Restoring IM pods to $REPLICAS replicas"
			$kubectl -n $NS scale --replicas=$REPLICAS sts isvgim
			log "$kubectl -n $NS scale --replicas=$REPLICAS sts isvgim"
		else
			printf "Restarting IM pod.\n"
			log "No replicas running previously, restarting single IM pod"
			$kubectl -n $NS scale --replicas=1 sts isvgim
			log "$kubectl -n $NS scale --replicas=1 sts isvgim"
		fi
	fi
	if [ $DRYRUN -eq 0 ]; then
		printf "\nFix installation of $TAG completed\n"
		log "Fix installation of $TAG completed."
	else
		printf "\nDry run complete.  Installation left as $CURLEVELNAME.\n"
		log "Dry run complete.  Installation left as $CURLEVELNAME."
	fi
	print_end
} #finish_installation

print_version() {
	IMAGE=$(kubectl -n $NS get sts -o wide | grep isvgim| awk '{ print $5 }' | cut -d ',' -f 1)
	install_version
	echo "Current installation version: $CURLEVELNAME"
	echo "Current image deployed: $IMAGE"
} #print_version

print_history() {
	printf "\n\nListing fix installation history\n\n"
	printf "DATE\t\tFIX NAME\tINSTALLER\n"
	if [ ! -f $HISTFILE ]; then
		echo "     ----- No fixes installed -----"
		printf "\n\n"
		exit 0
	fi
	sed 1d $HISTFILE | awk '{ print $3 "\t" $2 "\t" $4 }'
	printf "\n\n"
} #print_history

install_version() {
	if [ -f $HISTFILE ]; then
		CURLEVEL=$(tail -n 1 $HISTFILE | awk '{ print $1 }')
		CURLEVELNAME=$(tail -n 1 $HISTFILE | awk '{ print $2 }')
	else
		CURLEVEL=0
		CURLEVELNAME=10.0.2.0
	fi
} #install_version

# Make sure the script is copied into the correct directory
check_dir() {
	cd $(dirname ${BASH_SOURCE[0]})
	if [ ! -d ../helm ] && [ ! -d ../license ]; then
		echo "Please place the IMFixInstaller.sh in the starter/bin directory and run again."
		print_end
		exit 1
	fi
} #check_dir

# Make sure ISVGIM is installed
check_cluster() {
	printf "Checking for the IM cluster\n"
	log "Checking for the IM cluster"
	$kubectl get ns | grep -q $NS
	if [ $(echo $?) -ne 0 ]; then
		printf "\nThe IM Container must be installed into kubernetes before a fix\n"
		printf "can be applied.\n"
		log "The namespace specified in values.yaml ($NS) does not exist in kubernetes.  Aborting!"
		print_end
		exit 2
	fi
} #check_cluster

# Check for number of replicas running
check_replicas() {
	REPLICAS=$($kubectl -n $NS get sts isvgim -o custom-columns=":status.replicas" --no-headers)
	log "Found $REPLICAS IM pods running"
	if [ "x$REPLICAS" = "x" ]; then
		printf "\nUnable to find the isvgim stateful set in kubernetes.  Aborting!\n"
		log "$kubectl -n $NS get sts isvgim -o custom-columns=":status.replicas" --no-headers"
		log "Unable to find the isvgim stateful set in kubernetes.  Aborting!"
		print_end
		exit 3
	fi
} #check_replicas

# Check our version and what we're trying to install
check_version() {
	install_version
	echo "Found current installation version: $CURLEVELNAME."
	log "Found current installation version: $CURLEVELNAME."
	if [ "$TAG" = "$CURLEVELNAME" ]; then
		printf "\nNothing to do.  Installation is already up to date.\n"
		log "Current version is already at the new version level.  Exiting."
		print_end
		exit 0
	else
		printf "Will attempt an upgrade to version: $TAG.\n\n"
		log "Attempting to upgrade to $TAG"
	fi
} #check_version

check_prereqs() {
	NS=$(./sys/getNamespace.sh)
	SED=$(./sys/sedCheck.sh)
	kubectl=$(./sys/preReqCheck.sh)
	RC=($echo $?)
	if [ $RC -ne 0 ]; then
		echo "Unable to find a kubectl utility to access the cluster.  Aborting!"
		log "Unable to find usable kubectl.  Aborting!"
		print_end
		exit $RC
	fi
} #check_prereqs

do_prelim_checks() {
	check_dir
	check_prereqs
	check_version
	check_cluster
	check_replicas
} #do_prelim_checks

# Ask for permission to stop the cluster
stop_cluster() {
if [ $REPLICAS -ne 0 ]; then
	printf "Found $REPLICAS replica IM pods running. These must be stopped to apply the fix.\n\n"
	read -p "Would you like to stop them now? (yes/no) " YESNO
	if [ "x$YESNO" = "xyes" ]; then
		printf "\nScaling the cluster to 0\n"
		log "User agreed to scale IM pods to 0"
		$kubectl -n $NS scale --replicas=0 sts isvgim
		log "$kubectl -n $NS scale --replicas=0 sts isvgim"
	else
		printf "\nPlease rerun this utility when it is convenient to stop the cluster.\n"
		log "User declined to stop the IM servers.  Exiting."
		print_end
		exit 4
	fi
else
	echo "IM Cluster is already stopped, continuing"
	log "IM Server already stopped, continuing"
fi
} #stop_cluster

# Start the configuration pod
deploy_config_pod() {
	printf "\nDeploying configuration pod\n"
	log "Deploying the configuration pod"
	./updateYaml.sh 201-deployment-isvgimconfig.yaml

	POD=""
	TIMER=0
	RS=$($kubectl -n $NS get rs | grep isvgimconfig | awk '{ print $1 }')
	log "Waiting on replica set: $kubectl -n $NS get rs | grep isvgimconfig | awk '{ print \$1 }'"
	# Waiting for pod to be ready
	while [ "x$POD" = "x" ]; do
		sleep 5
		log "Waiting on pod: $kubectl -n $NS get pods | grep isvgimconfig | grep Running | awk '{ print \$1 }'"
		POD=$($kubectl -n $NS get pods | grep isvgimconfig | grep Running | awk '{ print $1 }')
		if [ $TIMER -gt 60 ]; then
			$kubectl -n $NS describe rs $RS
			printf "\nFailed to schedule the ISVGIM Config pod after 5 minutes.  Aborting!\n"
			log "pod failed to schedule after 5 minutes.  Aborting!"
			early_reset
			exit 5
		fi
		TIMER=$(($TIMER+1))
	done

	./sys/waitFor.sh $POD pod
	RC=$(echo $?)
	if [ $RC -ne 0 ]; then
		$kubectl -n $NS describe rs $RS
		printf "\nISVGIM Config pod is still not ready.  Aborting!\n"
		log "Config pod still not ready.  Aborting!"
		early_reset
		exit $RC
	fi
	./sys/waitFor.sh isvgimconfig application
	RC=$(echo $?)
	if [ $RC -ne 0 ]; then
		$kubectl -n $NS describe rs $RS
		printf "\nISVGIM Config failed to start.  Aborting!\n"
		log "Config pod failed to start.  Aborting!"
		early_reset
		exit $RC
	fi
} #deploy_config_pod

log() {
	TSTAMP=$(date +"%Y-%m-%d %H:%M:%S")
	echo "$TSTAMP $1" >> ../logs/${TAG}.log
} #log

# Make sure needed directories exist
setup_dirs() {
	if [ ! -d $FIXDIR ]; then
		log "Creating directory $FIXDIR"
		mkdir $FIXDIR
	fi
	if [ ! -d $BKUPDIR ]; then
		log "Creating directory $BKUPDIR"
		mkdir -p $BKUPDIR
	fi
	if [ ! -d $LOGDIR ]; then
		log "Creating directory $LOGDIR"
		mkdir $LOGDIR
	fi
} #setup_dirs

acquire_package() {
	echo "Downloading the Fix Package"
	log "Downloading the Fix Package"
	$kubectl -n $NS cp $POD:/work/$FIXPKG $FIXDIR/$FIXPKG > /dev/null 2>&1
	log "$kubectl -n $NS cp $POD:/work/$FIXPKG $FIXDIR/$FIXPKG > /dev/null 2>&1"
	if [ ! -f $FIXDIR/$FIXPKG ]; then
		printf "\nFailed to copy the Fix Package to the host. Aborting!\n"
		log "Failed to copy the Fix Package to the host.  Aborting!"
		early_reset
		exit 6
	fi
	verify_fix

	echo "Extracting the Fix Package"
	log "Extracting the Fix Package"
	RESULTS=$(tar zxf $FIXDIR/$FIXPKG -C $FIXDIR)
	if [ $(echo $?) -ne 0 ]; then
		printf "\nFailed to uncompress the fix package.  Aborting!\n"
		log "Failed to uncompress the fix package.  Aborting!"
		early_reset
		exit 7
	fi
	rm $FIXDIR/$FIXPKG
} #acquire_package

# Compare SHA256 hash values to confirm package is correct
verify_fix() {
	HASHCODE=$(kubectl -n $NS exec $POD -- bash -c "cat /work/fixPackage.CHECKSUM")
	log "Expected Hash Value: $HASHCODE"
	sha256sum --version > /dev/null 2>&1
	if [ $(echo $?) -ne 0 ]; then
		shasum --version > /dev/null 2>&1
		if [ $(echo $?) -ne 0 ]; then
			echo "No SHA-256 hash program found.  Unable to verify Fix Pacakge."
			log "Unable to locate a SHA-256 hash program."
			read -p "Would you like to continue anyway? (y/n) " YESNO
			if [ "x$YESNO" = "xy" ]; then
				log "User agreed to continue"
				return
			fi
		else
			SUMAPP="shasum -a 256"
		fi
	else
		SUMAPP="sha256sum"
	fi
	SUMVAL=$($SUMAPP ../fixes/fixPackage.tgz | awk '{ print $1 }')
	log "$SUMAPP ../fixes/fixPackage.tgz | awk '{ print \$1 }'"
	log "Calculated Hash value: $SUMVAL"
	if [ $SUMVAL = $HASHCODE ]; then
		echo "Fix Package verified for $TAG."
		log "Hash codes match!"
	else
		echo "The hash codes do not match:"
		log "Hash codes do NOT match. Aborting!"
		echo "  Expected: $HASHCODE"
		echo "  Computed: $SUMVAL"
		printf "\nThe file is either corrupt, or there is a version mismatch.\n"
		printf "Aborting the upgrade!\n\n"
		print_end
		exit 10
	fi	
} #verify_fix

apply_fix() {
	FIXNAME=$(cat $FIXDIR/$1/fixname)
	DATESTAMP=$(date +%Y-%m-%d)
	USER=$(whoami)
	printf "\nApplying fix $FIXNAME\n"
	log "Applying fix $FIXNAME"
	(cd $FIXDIR/$1/; ./applyFix.sh $DRYRUN)
	RC=$(echo $?)
	if [ $RC -ne 0 ]; then
		printf "\nError encountered!\nRolling back fix $FIXNAME.\n"
		log "Error installing $FIXNAME - Rolling back the fix"
		(cd $FIXDIR/$1/; ./applyFix.sh $DRYRUN rollback)
		log "(cd $FIXDIR/$1/; ./applyFix.sh $DRYRUN rollback)"
		cp $BKUPDIR/values.yaml $VALFILE
		log "cp $BKUPDIR/values.yaml $VALFILE"
		if [ "x$LASTGOOD" != "x" ]; then
			$SED "s;\(  isvgim: \)\(.*\):.*$;\1\2:$LASTGOOD;" $VALFILE
			log "Updated $VALFILE to use $LASTGOOD tag"
		fi
		printf "\nExiting with fixlevel at $LASTGOOD.\n"
		log "Exiting with fixlevel set to $LASTGOOD"
		finish_install
		exit 9
	fi
	if [ $DRYRUN -eq 0 ]; then
		printf "$1\t$FIXNAME\t$DATESTAMP\t$USER\n" >> $HISTFILE
		log "Updating $HISTFILE with: \"$1\t$FIXNAME\t$DATESTAMP\t$USER\n\""
		LASTGOOD=$FIXNAME
	fi
} #apply_fix

main() {
	log "**********                         **********"
	log "********** Starting IMFixInstaller **********"
	log "**********                         **********"
	do_prelim_checks
	stop_cluster
	setup_dirs

	# Save current image tag in case of rollback
	LASTGOOD=$(grep "  isvgim: " $VALFILE | cut -d ':' -f 3)

	# Update values.yaml with new tag for isvgim container
	cp $VALFILE $BKUPDIR/values.yaml
	log "Saving backup of values.yaml to $BKUPDIR/values.yaml"
	$SED "s;\(  isvgim: \)\(.*\):.*$;\1\2:$TAG;" $VALFILE
	log "$SED \"s;\(  isvgim: \)\(.*\):.*$;\1\2:$TAG;\" $VALFILE"

	deploy_config_pod
	acquire_package

	if [ $DRYRUN -eq 1 ]; then
		printf "\n***** Operating in Dry Run mode -- No changes will be made! *****\n"
		log "User enabled Dry Run mode -- no changes will be made!!"
	fi

	FIXES=$(ls $FIXDIR)
	for FIX in $FIXES; do
		if [ $FIX -gt $CURLEVEL ]; then
			apply_fix $FIX
		fi
	done

	if [ $DRYRUN -eq 1 ]; then
		restore_tag
	fi
	finish_install
} #main


# Start of main script
cd $(dirname ${BASH_SOURCE[0]})
case $(awk -vs1="$1" 'BEGIN { print tolower(s1) }') in
	-help|--help)
		print_help
		;;
	-dryrun|--dryrun)
		DRYRUN=1
		main
		;;
	-version|--version|-versions|--versions)
		check_prereqs
		print_version
		;;
	-history|--history)
		print_history
		;;
	*)
		main
		;;
esac

